import iced_x86 as ix86
import typing, enum, struct, json
from triton import TritonContext, Instruction, ARCH, REG

class PE(object):

    def __init__(self, filename: str):
        self._filename = filename

        with open(self._filename, "rb") as f:
            self._data = f.read()

        self._nt_header_offset = self.read_u32(0x3c)
        self._n_sections = self.read_u16(self._nt_header_offset + 0x6)
        self._n_sections_offset = self._nt_header_offset + self.read_u16(self._nt_header_offset + 0x14) + 0x18 
        self._image_base = self.read_u64(self._nt_header_offset + 0x30)
        self._sections = {}
        for i in range(self._n_sections):
            section_start_offset = self._n_sections_offset + 0x28 * i
            
            name = self._data[section_start_offset : section_start_offset + 0x8].rstrip(b"\x00").decode("iso-8859-1")
            va_size = self.read_u32(section_start_offset + 0x8 )
            rva = self.read_u32(section_start_offset + 0x0c )
            raw_size = self.read_u32(section_start_offset + 0x10 )
            raw_offset = self.read_u32(section_start_offset + 0x14 )

            self._sections[name] = dict(rva=rva, va_size=va_size, raw_size=raw_size, raw_offset=raw_offset)


    def read_u64(self, x: int) -> int:
        return struct.unpack("<Q", self._data[x:x+8])[0]

    def read_u32(self, x: int) -> int:
        return struct.unpack("<I", self._data[x:x+4])[0]

    def read_u16(self, x: int) -> int:
        return struct.unpack("<H", self._data[x:x+2])[0]

    def at_va(self, va: int) -> slice:
        rva = va - self.image_base
        for s in self.sections.values():
            if rva >= s["rva"] and rva < s["rva"] + s["va_size"]:
                return slice(rva - s["rva"] + s["raw_offset"], s["raw_size"] + s["raw_offset"])
        raise ValueError("RVA is out of range")

    def va_to_offset(self, va: int) -> int:
        rva = va - self.image_base
        for s in self.sections.values():
            if rva >= s["rva"] and rva < s["rva"] + s["va_size"]:
                return rva - s["rva"] + s["raw_offset"]
        raise ValueError("RVA is out of range")

    @property 
    def image_base(self) -> int:
        return self._image_base 

    @property 
    def sections(self) -> dict:
        return self._sections  

    @property 
    def data(self) -> bytes:
        return self._data 
    
    @property 
    def filename(self) -> str:
        return self._filename

class BasicBlockEnd(enum.Enum):
    CONDITIONAL = 1,
    UNCONDITIONAL = 2,
    INDIRECT = 3,
    SPLIT = 2,
    NOT_DONE = 4

class BasicBlock(object):
    def __init__(self, start_va: int, instructions: typing.List[ix86.Instruction], machine_code: typing.List[bytes]):
        self._start_va = start_va
        
        self._next = None 
        self._branch = None

        self._split = None
        self._simplified = None 
        self._label = 0

        self._set_instructions(instructions, machine_code)

        
    def _set_instructions(self, instructions: typing.List[ix86.Instruction], machine_code: typing.List[bytes]):
        self._instructions = instructions
        self._machine_code = machine_code
        self._update_end()
        self._update_size()

    
    def set_next(self, bb: "BasicBlock"):
        self._next = bb
    
    def set_branch(self, bb: "BasicBlock"):
        self._branch = bb

    def set_simplified(self, instructions: typing.List[ix86.Instruction]):
        self._simplified = instructions 

    @property 
    def simplified(self) -> typing.List[ix86.Instruction]:
        return self._simplified

    def _update_size(self):
        self._size = sum(map(lambda i: i.len, self._instructions)) 

    def _update_end(self):
        if not self._instructions:
            self._end = BasicBlockEnd.NOT_DONE
        elif self._instructions[-1].flow_control in [ix86.FlowControl.CONDITIONAL_BRANCH]:
            self._end = BasicBlockEnd.CONDITIONAL 
        elif self._instructions[-1].flow_control in [ix86.FlowControl.UNCONDITIONAL_BRANCH]:
            self._end = BasicBlockEnd.UNCONDITIONAL 
        elif self._instructions[-1].flow_control in [ix86.FlowControl.INDIRECT_BRANCH, ix86.FlowControl.RETURN]:
            self._end = BasicBlockEnd.INDIRECT  
        else:
            self._end = BasicBlockEnd.NOT_DONE

    def _add_instruction(self, instruction: ix86.Instruction, opcodes: bytes):
        self._instructions.append(instruction)
        self._machine_code.append(opcodes)
        self._update_end()
        self._update_size()   

    @property 
    def machine_code(self) -> typing.List[bytes]:
        return self._machine_code

    @property 
    def size(self) -> int:
        return self._size 

    @property 
    def instructions(self) -> typing.List[ix86.Instruction]:
        return self._instructions

    @property 
    def va(self) -> int:
        return self._start_va 

    @property 
    def end(self) -> BasicBlockEnd:
        return self._end 
    
    @property
    def next(self) -> "BasicBlock":
        return self._next 
    
    @property
    def branch(self) -> "BasicBlock":
        return self._branch 

    def encompass(self, va: int) -> bool:
        return va >= self.va and va < self.va + self.size

    def set_label(self, id: int):
        self._label = id 
    
    @property
    def label(self) -> int:
        return self._label

    

    def _begin_split(self, va: int) -> "BasicBlock":
        if not self.encompass(va):
            raise ValueError("VA is not inside this basic block")

        next_block = [(i, c) for i, c in zip(self._instructions, self._machine_code) if i.ip >= va]
        this_block = [(i, c) for i, c in zip(self._instructions, self._machine_code) if i.ip < va]

        if next_block[0][0].ip != va:
            raise ValueError("VA splits an instruction")
        next_block = BasicBlock(va, [i for i,_ in next_block], [c for _,c in next_block])
        next_block._branch = self._branch
        next_block._next = self._next 

        self._split = dict(
            instructions = [i for i,_ in this_block],
            machine_code = [c for _,c in this_block],
            branch = None,
            next = next_block)

        return next_block

    def _finish_split(self):
        if self._split:
            self._split["next"]._branch = self._branch
            self._split["next"]._next = self._next
            self._set_instructions(self._split["instructions"], self._split["machine_code"])
            self._branch = self._split["branch"]
            self._next = self._split["next"]
            self._end = BasicBlockEnd.SPLIT
            
            self._split = None 

    def __str__(self) -> str:
        return f"{hex(self.va)}: {('next ' + hex(self.next.va)) if self.next else ''} {('branch ' + hex(self.branch.va)) if self.branch else ''} ({self.size} bytes)"

    def __repr__(self) -> str:
        return hex(self.va)

    @classmethod
    def walk_function(cls, va: int, context: typing.Dict[int, "BasicBlock"], pe: PE) -> "BasicBlock":
        #print("Walk at: " + hex(va))
        if va in context:
            #print("BB already done")
            return context[va] 

        containing_bb = [bb for bb in context.values() if bb.encompass(va)]
        if containing_bb:
            #print(f"Splitting bb at {hex(containing_bb[0].va)}")
            splitted = containing_bb[0]._begin_split(va)
            context[va] = splitted
            return splitted

        bb = BasicBlock(va, [], [])
        context[va] = bb
        #print("Ready " + str(len(pe[at_va(va)])))
        code = pe.data[pe.at_va(va)]
        offset = 0
        decoder = ix86.Decoder(64, code, ip=va)
        for i in decoder:
            bb._add_instruction(i, code[offset:offset+i.len])
            offset += i.len

            #print(bb.end)
            if bb.end == BasicBlockEnd.CONDITIONAL:
                next_bb = cls.walk_function(i.next_ip, context, pe)
                if bb._next is None:
                    bb._next = next_bb
                bb._branch = cls.walk_function(i.near_branch_target, context, pe)
                bb._finish_split()
                return bb 
            if bb.end == BasicBlockEnd.UNCONDITIONAL:
                bb._next = cls.walk_function(i.near_branch_target, context, pe)
                bb._finish_split()
                return bb 
            if bb.end == BasicBlockEnd.INDIRECT:
                bb._finish_split()
                return bb 


"""
    Represent and encoded string
"""
class EncodedString(object):

    """
        decoding_start_ip: The VA where the decoding code starts
        post_decoding_ip: The VA of the first instruction AFTER the decoding code

        encoded_string: The bytes of the encoded string
        first_char_disp: The offset (from RBP/RSP, doesn't matter) of the first char of the string. This is needed for the concolic execution
        flag_disp: The offset of the string's flag. Not really needed but useful just in case.
        patch_chars: The VA of the immediate bytes of the store instructions that generate the string (useful to patch them).
        patch_jump: The VA of the jnz instruction that jump on the string's flag (useful to patch it)
        first_char_ip: The VA of the store that stores the first char of the string (useful for debugging purpose)
        pe: a PE object for patching the original code
    """
    def __init__(self, decoding_start_ip: int, post_decoding_ip: int, encoded_string: bytearray, first_char_disp: int, flag_disp: int, patch_chars: list, patch_jump:int, first_char_ip: int, pe: PE):
        self._decoding_start_ip = decoding_start_ip
        self._post_decoding_ip = post_decoding_ip
        self._encoded_string = encoded_string
        self._first_char_disp = first_char_disp
        self._flag_disp = flag_disp
        self._pe = pe 
        self._decoded = None
        self._patch_chars = patch_chars
        self._patch_jump = patch_jump
        self._first_char_ip = first_char_ip

    """
        Get the encoded string
    """
    @property
    def encoded(self) -> bytearray:
        return self._encoded_string

    """
        Get the decoded string
        The result is cached
    """
    @property
    def decoded(self) -> bytearray:
        if self._decoded is None:

            #Create a Triton context
            ctx = TritonContext()
            #For the X86-64 architecture 
            ctx.setArchitecture(ARCH.X86_64)

            #Since this is a simulation we can make RBP/RSP point to zero so we can use the offset given directly without adding any base
            ctx.setConcreteRegisterValue(ctx.registers.rsp, 0)
            ctx.setConcreteRegisterValue(ctx.registers.rbp, 0)
            
            #We start the execution from the decoding code
            ctx.setConcreteRegisterValue(ctx.registers.rip, self._decoding_start_ip)

            #Set the encoded string in memory
            ctx.setConcreteMemoryAreaValue(self._first_char_disp, self._encoded_string)
            #Set the flag also (though it's not read)
            ctx.setConcreteMemoryAreaValue(self._flag_disp, bytes([0]))

            #We continue execution until we reach the first instruction after the decoding code
            while ctx.getConcreteRegisterValue(ctx.registers.rip) != self._post_decoding_ip:
                #Get current RIP value
                ip = ctx.getConcreteRegisterValue(ctx.registers.rip)

                #Make a new instruction (Triton will execute a single instruction regardless of how many bytes are set with setOpcode, nice!)
                cur_ins = Instruction()
                cur_ins.setAddress(ip)
                cur_ins.setOpcode(self._pe.data[self._pe.at_va(ip)][:16])

                #Single step of concolic execution!
                ctx.processing(cur_ins)

            #OK, time to read back from what was once the encoded string
            self._decoded = ctx.getConcreteMemoryAreaValue(self._first_char_disp, len(self._encoded_string))

        return self._decoded

    """
        Return a dictionary of <VA>:<BYTE> that represent a patch.
        Each entry tells what byte to write at what address

        The patch makes the original stores store the decoded bytes instead of the encoded one.
        It also patch the jnz jump into a jmp to simplify the IDA graph.
    """
    @property 
    def patch(self) -> dict[int, bytes]:
        #Get the decoded string and do a sanity check
        sd = self.decoded 
        assert(len(sd) == len(self._patch_chars))

        #Patch the stores bytes
        ptc = {a:v for a,v in zip(self._patch_chars, sd)}

        #Patch the jnz (long and short form)
        opcode = self._pe.data[self._pe.at_va(self._patch_jump)][0]
        if opcode == 0x0f:
            ptc[self._patch_jump] = 0x90
            ptc[self._patch_jump + 1] = 0xe9
        else:
            ptc[self._patch_jump] = 0xeb

        return ptc 

    @property 
    def first_char_ip(self) -> int:
        return self._first_char_ip

    """
        Given the data of the original PE, apply the self.patch path to it
    """
    def patch_pe(self, data):
        for a, v in self.patch.items():
            o = self._pe.va_to_offset(a)
            data[o] = v 
        

"""
    Check if the BasicBlock bb contains an encoded string.
    The code handling the string is necessarily at the end of the BB if present.
"""
def is_encoded_string(bb: BasicBlock, pe: PE) -> typing.Union[EncodedString, False]:
    """
        Check if an instruction has the form <mnemonic> <reg>, ...
    """
    def is_op_reg(i, mnemonic, dst_reg):
        return i.mnemonic == mnemonic and i.op0_kind == ix86.OpKind.REGISTER and (i.op0_register in dst_reg if isinstance(dst_reg, list) else (i.op0_register == dst_reg))
    """
        Get the displacemente from RSP/RBP in loads from memory (e.g.: in mov eax, [rbp+241] this function returns 241). The the instruction is not a load
        from rbp/rsp the returned value is None
    """
    def get_rsp_disp(i):
        return i.memory_displacement if i.op1_kind == ix86.OpKind.MEMORY and i.memory_base in [ix86.Register.RSP, ix86.Register.RBP] and i.memory_index == ix86.Register.NONE else None
    """
        Check if the source of the instruction is an immediate
    """
    def is_src_imm(i, imm):
        return i.op1_kind == ix86.OpKind.IMMEDIATE8 and i.immediate(1) == imm
    
    """
        Get the displacement from RBP/RSP and the value of a store. (e.g. in mov [rbp+241], 56 this functions return (241, 56)).
        If the instruction is not a store to RBP/RSP, (None, None is returned)
    """
    def get_mov_rsp_off_imm(i):
        return (i.memory_displacement, i.immediate(1)) if (
            i.mnemonic == ix86.Mnemonic.MOV and i.op0_kind == ix86.OpKind.MEMORY and i.memory_base in [ix86.Register.RSP, ix86.Register.RBP] and i.memory_index == ix86.Register.NONE 
            and i.op1_kind == ix86.OpKind.IMMEDIATE8) else (None, None)


    """
        Step 1. Check if the BB end with three specific instruction.
    """

    #Three instructions are the "signature" of string decoding code, one if for storing the flag and one is for storing at least one char of the string
    if len(bb.instructions) < 5:
        return False 

    #
    #Last instructions:
    #

    # must be mov al, [rsp/rbp+x] or mov eax, [rsp/rbp+x]
    # set string_start_disp to be the offset in the stack (from rbp/rsp, it doesn't matter) of the string's flag (which is just before the beginning of the string)
    if (not (
            is_op_reg(bb.instructions[-3], ix86.Mnemonic.MOV, [ix86.Register.AL, ix86.Register.DL, ix86.Register.CL]) 
            or 
            is_op_reg(bb.instructions[-3], ix86.Mnemonic.MOVZX, ix86.Register.EAX)
    )
        or 
        (string_start_disp := get_rsp_disp(bb.instructions[-3])) is None):
        return False 
  
    # test al, 1; jnz 
    if (not is_op_reg(bb.instructions[-2], ix86.Mnemonic.TEST, [ix86.Register.AL, ix86.Register.DL, ix86.Register.CL]) 
        or not is_src_imm(bb.instructions[-2], 1)):
        return False 

    #jnz 
    if bb.instructions[-1].mnemonic != ix86.Mnemonic.JNE:
        return False 

    #This is the VA of the first instruction after the code that decode the string. It's the target of the jnz instruction (that conveniently)
    post_decoding_ip = bb.instructions[-1].near_branch_target
    #This is the VA of the first instruction of the decoding code
    decoding_ip = bb.instructions[-3].ip
    #This is the VA of the jnz instruction itself (useful to patch it)
    patch_jump = bb.instructions[-1].ip 

    #The bytes of the encoded string
    encoded_string = []
    #The offset from RSP/RBP of the first byte of the string (this is just string_start_disp + 1 anyway)
    first_char_disp = 0xffffffff
    #The address of the immediate bytes of the stores that generate the encoded string in the stack (useful to patch them)
    patch_chars = []
    #
    first_char_ip = None

    #Scan each instruction in the BB in reverse order (skipping the last three)
    for i in range(len(bb.instructions)-4, -1, -1):
        ins = bb.instructions[i]

        #If this is a store to RSP/RBP of a byte, get the offset and the byte
        disp, imm = get_mov_rsp_off_imm(ins)

        #Not a store, ignore this instruction. The compiler probably reordered it
        if disp is None:
            continue 

        #If this is the store that stores the flag, we have reached back far enough, return an EncodedString
        if disp == string_start_disp:
            #Just some sanity check
            if imm != 0 or first_char_ip is None:
                return False 
            
            #We are done, return an object with all the info
            return EncodedString(decoding_ip, post_decoding_ip, bytearray(encoded_string), first_char_disp, string_start_disp, patch_chars, patch_jump, first_char_ip, pe)
        
        #This could be a store for an encoded char then
        else:

            #Update the minimum displacement seen so far (and the VA of the store)
            if disp < first_char_disp:
                first_char_disp = min(first_char_disp, disp)
                first_char_ip = ins.ip 

            #Add the encoded char at the beginning (we are scanning backwards)
            encoded_string.insert(0, imm)

            #We need the offset of the immediate byte in the store, is there a better method than this?
            decoder = ix86.Decoder(64, bytes(), ip=ins.ip)
            offsets = decoder.get_constant_offsets(ins)
            patch_chars.insert(0, ins.ip+offsets.immediate_offset)

    #Something went wrong
    return False

"""
    Given the first BB of a function, find all the encoded strings
"""
def find_encoded_strings(bb: BasicBlock, pe: PE) -> typing.Union[EncodedString, False]:
    """
        Each BB is labelled with 1 once visited to avoid infinite recursion
    """

    #Is this BB already visited?
    if bb.label != 0:
        return []

    #Mark this BB as visited
    bb.set_label(1)

    #Encoded strings found
    strings = []

    #If this BB decodes a string, add it to the list
    res = is_encoded_string(bb, pe)
    if res:
        strings.append(res)

    #Recurse to the next BB
    if bb.next:
        strings += find_encoded_strings(bb.next, pe)

    #Recurse to the branch BB
    if bb.branch:
        strings += find_encoded_strings(bb.branch, pe)

    return strings 



"""
    ENTRY-POINT
"""


#Read the sample (edit the path)
sample = PE("knight.exe")

#All the functions identified by IDA in the malware (excluding the runtime ones)
functions = [
    0x14000f420, 0x140001000, 0x1400015C0, 0x140001720, 0x1400036C0, 0x140003810, 0x140003970, 0x140003F60, 0x140004140, 0x140004530, 0x1400046A0, 0x140004720, 0x1400048A0, 0x140004960, 0x140004970, 0x140004A60, 0x140004A70, 0x140004B30, 0x140004C20, 0x140004D20, 0x140009100, 0x140009190, 0x140009690, 0x1400096A0, 0x1400096B0, 0x1400096C0, 0x1400097F0, 0x140009880, 0x14000A6F0, 0x14000A9D0, 0x14000AA70, 0x14000AA80, 0x14000ABC0, 0x14000C200, 0x14000C460, 0x14000C4C0, 0x14000E390, 0x14000ECB0, 0x14000ED10, 0x14000EEC0, 0x140013560, 0x140013890, 0x1400138C0, 0x1400138F0, 0x140013AD0, 0x140013B00, 0x140014320, 0x140016830, 0x140016AF0, 0x140016C10, 0x140016E00, 0x140016F10, 0x14001AA60, 0x140023830, 0x140023FE0, 0x1400244B0, 0x1400248D0, 0x140024A70, 0x140024F10, 0x140025390, 0x1400257C0, 0x140025C60, 0x1400260B0, 0x140026530, 0x1400269B0, 0x140026E30, 0x1400272B0, 0x1400276F0, 0x140027B70, 0x140028020, 0x140028520, 0x1400289A0, 0x140028E70, 0x1400292C0, 0x140029710, 0x140029B50, 0x140029FD0, 0x14002A450, 0x14002A920, 0x14002ADC0, 0x14002B240, 0x14002B6E0, 0x14002BB30, 0x14002BFB0, 0x14002C430, 0x14002C8B0, 0x14002CD50, 0x14002D1D0, 0x14002D650, 0x14002DB10, 0x14002DF90, 0x14002E440, 0x14002E910, 0x14002ED60, 0x14002F260, 0x14002F6E0, 0x14002FBB0, 0x140030000, 0x140030440, 0x1400308C0, 0x140030D40, 0x1400314F0, 0x1400315C0, 0x1400316C0, 0x1400318A0, 0x140031AB0, 0x140031B70, 0x140031DB0, 0x140031F00, 0x140031F34, 0x140031F58, 0x140032020, 0x140032040, 0x140032060, 0x140032080, 0x140032778, 0x140032788, 0x1400327AC, 0x1400327C0, 0x1400327C8, 0x1400327D0, 0x140032880, 0x140032A28, 0x140032A64, 0x1400342C8, 0x140034314, 0x140034344, 0x140034438, 0x140034440, 0x1400344C0, 0x140034560, 0x1400345C8, 0x140034838, 0x140034A40, 0x140034AB4, 0x140034C00, 0x140034D70, 0x140034FD8, 0x1400350A0, 0x140035368, 0x1400354E8, 0x140035570, 0x1400358BC, 0x1400358F8, 0x140035A70, 0x140035B9C, 0x140035E74, 0x140035ED4, 0x140036074, 0x140036120, 0x140036130, 0x1400362E0, 0x14003656C, 0x140036588, 0x140036994, 0x140036A4C, 0x140036BD0, 0x140036BD8, 0x140036C20, 0x140036C54, 0x14003756C, 0x1400375DC, 0x140037AA4, 0x140038078, 0x140038338, 0x1400388D4, 0x140039210, 0x14003922C, 0x1400392D0, 0x14003944C, 0x1400399FC, 0x140039BAC, 0x14003A650, 0x14003A6DC, 0x14003A998, 0x14003AC3C, 0x14003AC48, 0x14003AF08, 0x14003AF80, 0x14003AFAC, 0x14003B2B0, 0x14003B438, 0x14003B5D4, 0x14003B608, 0x14003B63C, 0x14003BC38, 0x14003BD50, 0x14003C238, 0x14003C67C, 0x14003C8C8, 0x14003CFF8, 0x14003D000, 0x14003D3CC, 0x14003D604, 0x14003D888, 0x14003D930, 0x14003D980, 0x14003DA28, 0x14003DA50, 0x14003DC2C, 0x14003DF20, 0x14003E4B8, 0x14003E5A4, 0x14003E888, 0x14003F1B8, 0x14003F258, 0x14003F360, 0x14003F560, 0x14003F614, 0x14003F740, 0x14003F924, 0x140040B70, 0x140040FB0, 0x1400413A0, 0x140041514, 0x140041550, 0x140041638, 0x140041894, 0x1400418E0, 0x140041FAC, 0x140042084, 0x1400423FC, 0x1400424A0, 
]




"""
    Emulate a piece of code and read back from memory

    start, end: The starting and end VAs of the code to execute
    offset, size: The area of memory to read back once the execution ends

    RSP and RBP are set to zero before executing the code
"""
def emulate(start, end, offset, size):
    ctx = TritonContext()
    ctx.setArchitecture(ARCH.X86_64)

    ctx.setConcreteRegisterValue(ctx.registers.rsp, 0)
    ctx.setConcreteRegisterValue(ctx.registers.rsp, 0)
    ctx.setConcreteRegisterValue(ctx.registers.rip, start)

    while ctx.getConcreteRegisterValue(ctx.registers.rip) != end:
        cur_ins = Instruction()
        ip = ctx.getConcreteRegisterValue(ctx.registers.rip)
        cur_ins.setAddress(ip)
        cur_ins.setOpcode(sample.data[sample.at_va(ip)][:16])
        ctx.processing(cur_ins)

    return ctx.getConcreteMemoryAreaValue(offset, size)

"""
    Patch the PE in output_pe and print an IDC script for adding a comment for each string with the decoded values
"""
def patch_pe_and_idc(output_pe: str):
    #Make a copy of the PE data
    sample_data = bytearray([x for x in sample.data])

    #Make the IDC script (we cannot affort IDA Pro :( )
    idc = "static main() {"
    for f in functions:
        ctx = {}
        bb = BasicBlock.walk_function(f, ctx, sample)
        strings = find_encoded_strings(bb, sample)
        for s in strings:
            idc += "\n\n"

            #Set the comment
            dec = json.dumps(s.decoded.replace(b'\x00', b'').decode("ISO-8859-1"))
            idc += f"set_cmt({hex(s.first_char_ip)}, {dec}, 0);\n"

            #Patch the PE
            s.patch_pe(sample_data)

    idc += "\n}\n"

    #Print the the script
    print(idc)

    #Write the patched PE
    with open(output_pe, "wb") as f:
        f.write(sample_data)

"""
    Decode all the string in the function starting at the given VA
"""
def single_function(va):
    ctx = {}
    bb = BasicBlock.walk_function(va, ctx, sample)
    strings = find_encoded_strings(bb, sample)
    for s in strings:
        print(s.decoded.replace(b'\x00', b'').decode("ISO-8859-1"))

"""
    Decode the string in the BB at the given VA
"""
def single_block(va):
    ctx = {}
    bb = BasicBlock.walk_function(va, ctx, sample)
    res = is_encoded_string(bb, sample)
    if res:
        #Remove null bytes and decode as ISO since any byte sequence is a valid encoding in ISO
        print(res.decoded.replace(b'\x00', b'').decode("ISO-8859-1"))




"""
    Examples:

    Find "all" string:
        patch_pe_and_idc("knight-patched.exe")

    Emulate a piece of code and read back from memory (useful to manually decode a string):
        print(emulate(0x1400165c0, 0x140016774, 0, 0x1054).decode("ISO-8859-1"))

    Decode a string given a BB start address:
        print(single_block(0x14000f420))

    Decode all strings in a function:
        print(single_function(0x14000f420))
"""


